/*                                              -*- linux-c -*-
 * rex5055d.c - RATOC REX-5055 DIO PC driver for Linux
 *
 * 2002-01-18
 *
 *
 * 2001-2002 (C) RATOC Systems, Inc.
 *
 */

/*
   This driver is for the RATOC REX-5055, a DIO PC Card by RATOC
   Systems, Inc. (http://www.ratocsystems.com/).
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <asm/system.h>

#define KERNEL
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>		/* for verify_area */
#include <linux/errno.h>	/* for -EBUSY */
#include <linux/ioport.h>	/* for check_region, request_region */
#include <linux/delay.h>	/* for loops_per_jiffy */
#include <asm/segment.h>	/* for put_user_byte */
#include <asm/io.h>		/* for inb_p, outb_p, inb, outb, etc. */
#include <asm/uaccess.h>	/* for get_user, etc. */
#include <asm/current.h>    /* for current */
#include <linux/wait.h>		/* for wait_queue */
#include <linux/init.h>		/* for __init, module_{init,exit} */
#include <linux/poll.h>		/* for POLLIN, etc. */
#include <linux/devfs_fs_kernel.h>
#include <linux/smp_lock.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ciscode.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>

#include "rex5055d.h"        /* local header file for REX-5055 values */

#define PCMCIA_DEBUG    1

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
static char *version =
"rex5055d.c 0.58 2002/01/18 (c) RATOC Systems, Inc.";
#else
#define DEBUG(n, args...)
#endif
/* for debug */
#define DebugDump(PTR, LEN)  { int i; for (i = 0; i<LEN; i++) printk("%02X ", PTR[i]); }

/* Parameters that can be set with 'insmod' */

/* Bit map of interrupts to choose from */
static u_int irq_mask = 0xdeb8;
static int irq_list[4] = { -1 };

MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");

/*====================================================================*/

typedef struct rex5055_info_t {
    dev_link_t	link;
    int         ndev;
    dev_node_t	node[4];
    int         Socket;
} rex5055_info_t;

typedef struct rexdio_profile_s {
    int    socket;    /* PCMCIA socket no. */
    int    pid;     /* Process ID */
    int    port;
    int    irq;
    char   *cmdbuf;
    int    major;
    devfs_handle_t  handle;
    int    (*in_func)(struct rexdio_profile_s *mpv, char *buf);
    int    bit;    /* for GetBit, SetBitOn, SetBitOff */
    int    ctrl;   /* control byte */
    int    stat;   /* IRQ status byte */
    int    dtype;  /* Data type - 0x03:Word  0x02:High  0x01:Low */
    int    dat;    /* Data */
    int    res;    /* response data */
} rexdio_profile;

#define  MPMAX   4
rexdio_profile  mp[MPMAX];     /* My Profile */
int  mp_socket = 0;               /* Number of mp */


static int rex5055_major;

/* prototypes for file_operations struct */

static dev_link_t *rex5055_attach(void);
static void rex5055_detach(dev_link_t *link);

static ssize_t rex5055_read(struct file *, char *,
			 size_t nbytes, loff_t * ppos);
static ssize_t rex5055_write(struct file *, const char *,
			  size_t nbytes, loff_t * ppos);
static unsigned int rex5055_poll(struct file *, poll_table *);
static int rex5055_open(struct inode *, struct file *);
static int rex5055_ioctl(struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg);
static int rex5055_close(struct inode *inode, struct file *file);

static void  rex5055_interrupt(int irq, void *dev_id, struct pt_regs *REG);

static void rex5055_release(u_long arg);
static void rex5055_config(dev_link_t *link);

static int rex5055_event(event_t event, int priority,
			event_callback_args_t *args);

static struct file_operations rex5055_fops =
{
	owner:		THIS_MODULE,
	read:		rex5055_read,
	write:		rex5055_write,
	poll:		rex5055_poll,
	ioctl:		rex5055_ioctl,
	open:       rex5055_open,
        release:    rex5055_close,
};

/* local prototypes */
static int rex5055_dev_probe(void);

static dev_info_t dev_info = "rex5055_cs";

static dev_link_t *dev_list = NULL;

static devfs_handle_t devfs_handle;

/*====================================================================*/
/*
 * from "drivers/pcmcia/cs_internal.h"
 * for refer of client->Socket
 */
typedef struct erase_busy_t {
    eraseq_entry_t	*erase;
    client_handle_t	client;
    struct timer_list	timeout;
    struct erase_busy_t	*prev, *next;
} erase_busy_t;

typedef struct client_t {
    u_short		client_magic;
    socket_t		Socket;
    u_char		Function;
    dev_info_t		dev_info;
    u_int		Attributes;
    u_int		state;
    event_t		EventMask, PendingEvents;
    int (*event_handler)(event_t event, int priority,
			 event_callback_args_t *);
    event_callback_args_t event_callback_args;
    struct client_t 	*next;
    u_int		mtd_count;
    wait_queue_head_t	mtd_req;
    erase_busy_t	erase_busy;
} client_t;

/*====================================================================*/

typedef struct rexdio_cmd_f {
      unsigned char *cmd;
      int  (*func)(rexdio_profile *mpv, char *buf);
} rexdio_cmd_f  ;

// static  char *cmd;

static int  rex5055_WriteCtrl(rexdio_profile *mpv, char *buf);
static int  rex5055_WriteMask(rexdio_profile *mpv, char *buf);
static int  rex5055_ReadMask(rexdio_profile *mpv, char *buf);
static int  rex5055_ReadLowByte(rexdio_profile *mpv, char *buf);
static int  rex5055_ReadHighByte(rexdio_profile *mpv, char *buf);
static int  rex5055_ReadWord(rexdio_profile *mpv, char *buf);
static int  rex5055_GetBit(rexdio_profile *mpv, char *buf);
static int  rex5055_WriteLowByte(rexdio_profile *mpv, char *buf);
static int  rex5055_WriteHighByte(rexdio_profile *mpv, char *buf);
static int  rex5055_WriteWord(rexdio_profile *mpv, char *buf);
static int  rex5055_SetBitOn(rexdio_profile *mpv, char *buf);
static int  rex5055_SetBitOff(rexdio_profile *mpv, char *buf);
static int  rex5055_PollIrqStat(rexdio_profile *mpv, char *buf);
static int  rex5055_SetIrqResponse(rexdio_profile *mpv, char *buf);

static int  doReadMask(rexdio_profile *mpv, char *buf);
static int  doReadLowByte(rexdio_profile *mpv, char *buf);
static int  doReadHighByte(rexdio_profile *mpv, char *buf);
static int  doReadWord(rexdio_profile *mpv, char *buf);
static int  doGetBit(rexdio_profile *mpv, char *buf);
static int  doIrqStat(rexdio_profile *mpv, char *buf);

/*====================================================================*/

typedef struct rexdio_ioctl_s {
      unsigned int  cmd;
      int  (*func)(rexdio_profile *mpv, unsigned long arg);
} rexdio_ioctl_f  ;

                                /* REXDIO_WRTCTRL 0xD001: Set Control Reg. */
static int ioctl5055_SetCtrl(rexdio_profile *mpv, unsigned long arg);
                                /* REXDIO_SETMASK 0xD002: Set Mask Reg. */
static int ioctl5055_SetMask(rexdio_profile *mpv, unsigned long arg);
                                /* REXDIO_GETMASK 0xD003: Set Mask Reg. */
static int ioctl5055_GetMask(rexdio_profile *mpv, unsigned long arg);
                                /* REXDIO_OUT_HIGHBYTE 0xD010: Out high byte (8-15) */
static int ioctl5055_OutHighByte(rexdio_profile *mpv, unsigned long arg);
                                /* REXDIO_IN_HIGHBYTE 0xD011: In high byte (8-15) */
static int ioctl5055_InHighByte(rexdio_profile *mpv, unsigned long arg);
                                /* REXDIO_OUT_LOWBYTE 0xD012: Out low byte (0-7) */
static int ioctl5055_OutLowByte(rexdio_profile *mpv, unsigned long arg);
                                /* REXDIO_IN_LOWBYTE 0xD013: In low byte (0-7) */
static int ioctl5055_InLowByte(rexdio_profile *mpv, unsigned long arg);
                                /* REXDIO_OUT_WORD 0xD014: Out word */
static int ioctl5055_OutWord(rexdio_profile *mpv, unsigned long arg);
                                /* REXDIO_IN_WORD 0xD015: In word */
static int ioctl5055_InWord(rexdio_profile *mpv, unsigned long arg);

                                /* REXDIO_IRQSTAT 0xD020: Irq status */
static int ioctl5055_IrqStat(rexdio_profile *mpv, unsigned long arg);
                      /* REXDIO_IRQRESPONSE 0xD021: Irq response */
static int ioctl5055_IrqResponse(rexdio_profile *mpv, unsigned long arg);

static int ioctl5055_DispatchCmd(rexdio_profile *mpv,
		      unsigned int cmd,
		      unsigned long arg);

/*====================================================================*/

static char *
SkipBlank(char *buf)
{
    char *p;

    for (p = buf; *p && (*p == ' ') ; p++)
        ;
    return p;
}

static int
HexToBin(char *buf, int *val)
{
    int work;
    char *p;

    p = SkipBlank(buf);    /* Skip blank char. */

    if (*p != '0' && *p != 'X')
        return -1;

    work = 0;
    for (p += 2; *p; p++)  {

        if ( *p < '0'  && *p > 'f')
            break;
        if ( *p >= '0' && *p <= '9')   {
            work = work << 4;
            work += (*p - '0');
        } else if ( *p >= 'A' && *p <= 'F')   {
            work = work << 4;
            work += (*p- 'A' + 10);
        } else if ( *p >= 'a' && *p <= 'f')   {
            work = work << 4;
            work += (*p- 'a' + 10);
        } else {
            break;
        }
    }

    *val = work;

    return (int)(p - buf);

}  /* HexToBin */

static int
DecToBin(char *buf, int *val)
{
    int work;
    char *p;

    p = SkipBlank(buf);    /* Skip blank char. */

    work = 0;
    for (  ; *p; p++)  {
        if (*p < '0' || *p > '9')
            break;
        work = work * 10;
        work += (*p - '0');
    }

    *val = work;

    return (int)(p-buf);

}  /* DecToBin */

/*====================================================================*/

/* Write command table */

rexdio_cmd_f   rex_cmd_tb[] = {
        { "WrtCtrl",  rex5055_WriteCtrl },
        { "WrtMask", rex5055_WriteMask },
        { "ReadMask", rex5055_ReadMask },
        { "ReadLowByte", rex5055_ReadLowByte },
        { "ReadHighByte", rex5055_ReadHighByte },
        { "ReadWord", rex5055_ReadWord },
        { "GetBit", rex5055_GetBit },
        { "WriteLowByte", rex5055_WriteLowByte },
        { "WriteHighByte", rex5055_WriteHighByte },
        { "WriteWord", rex5055_WriteWord },
        { "SetBitOn", rex5055_SetBitOn },
        { "SetBitOff", rex5055_SetBitOff },
        { "PollIrqStat", rex5055_PollIrqStat },
        { "SetIrqResponse", rex5055_SetIrqResponse },
        { NULL, NULL }
} ;

static int
rex5055_DispatchCmd(rexdio_profile *mpv, char *buf)
{
    int   i;
    int   len;
    char  *ptr, *nxt;


    ptr = strtok(buf,  ":");
    if (ptr == NULL) {
        printk("Delimiter not found: %02X: %s\n", buf[0], buf);
        return -1;      /* delimiter not found */
    }
    nxt = buf + strlen(buf) +1;

    for (i = 0; rex_cmd_tb[i].cmd != NULL; i++ ) {
        if ( !strcmp(buf, rex_cmd_tb[i].cmd) ) {
                break;
        }
    }
    if (rex_cmd_tb[i].cmd == NULL) {
        printk("Unknown command: %s\n", buf);
        return -1;
    }


//    printk("Cmd: %s Arg: %s\n", rex_cmd_tb[i].cmd, nxt);
    len = (rex_cmd_tb[i].func)(mpv, nxt);
    if (len == -1)
      return -1;

    nxt += len;

//  printk("Cmd: %s ret: %d\n", rex_cmd_tb[i].cmd, (int)(nxt-buf));

    return (int)(nxt-buf);

}  /* rex5055_DispatchCmd */

/*====================================================================*/

/*
 *       "WrtCtrl:0x%02\n"
 */

static int
rex5055_WriteCtrl(rexdio_profile *mpv, char *buf)
{
    int   val;
    int   len;

//    printk("rex5055_WriteCtrl:\n");
    len = HexToBin(buf, &val);
//    printk("Value: %02X\n", val);

    if (len != -1) {
      outb(val, mpv->port);
      mpv->ctrl = val;
    }
    mpv->in_func = NULL;

    return len;

}  /* rex5055_WriteCtrl */

/*
 *      "WrtMask:0x%02\n"
 */

static int
rex5055_WriteMask(rexdio_profile *mpv, char *buf)
{
    int  val;
    int  len;

//    printk("rex5055_WriteMask:\n");
    len = HexToBin(buf, &val);
//    printk("Value: %02X\n", val);
    if (len !=-1)
      outb(val, mpv->port+1);
    mpv->in_func = NULL;

    return len;

}  /* rex5055_WriteMask */

/*
 *       "ReadMask:\n"
 */

static int
rex5055_ReadMask(rexdio_profile *mpv, char *buf)
{

//    printk("rex5055_ReadMask:\n");
    mpv->in_func = doReadMask;
    return 0;
}

/*
 *       "ReadLowByte:\n"
 */

static int
rex5055_ReadLowByte(rexdio_profile *mpv, char *buf)
{
//    printk("rex5055_ReadLowByte:\n");

    mpv->in_func = doReadLowByte;
    return 0;
}

/*
 *       "ReadHighByte:\n"
 */

static int
rex5055_ReadHighByte(rexdio_profile *mpv, char *buf)
{
//    printk("rex5055_ReadHighByte:\n");

    mpv->in_func = doReadHighByte;
    return 0;
}

/*
 *       "ReadWord:\n"
 */

static int
rex5055_ReadWord(rexdio_profile *mpv, char *buf)
{
//    printk("rex5055_ReadWord:\n");

    mpv->in_func = doReadWord;
    return 0;
}

/*
 *       "getBit:\n"
 */

static int
rex5055_GetBit(rexdio_profile *mpv, char *buf)
{
    int  val;
    int  len;

//    printk("rex5055_GetBit:\n");
    len = DecToBin(buf, &val);
//    printk("Value: %d\n", val);
    if (len != -1) {
      mpv->in_func = doGetBit;
      mpv->bit = val;
    }

    return len;
}

/*
 *       "WriteLowByte:0xDD\n"
 */

static int
rex5055_WriteLowByte(rexdio_profile *mpv, char *buf)
{
    int  val;
    int  len;

//    printk("rex5055_WriteLowByte:\n");
    len = HexToBin(buf, &val);
//    printk("Value: %02X\n", val);

    if (len != -1)
      outb(val, mpv->port+2);
    mpv->in_func = NULL;

    return len;
}

/*
 *       "WriteHighByte:0xDD\n"
 */

static int
rex5055_WriteHighByte(rexdio_profile *mpv, char *buf)
{
    int  val;
    int  len;

//    printk("rex5055_WriteHighByte:\n");
    len = HexToBin(buf, &val);
//    printk("Value: %02X\n", val);

    if (len != -1)
      outb(val, mpv->port+3);
    mpv->in_func = NULL;

    return len;
}

/*
 *       "WriteWord:0xDDDD\n"
 */

static int
rex5055_WriteWord(rexdio_profile *mpv, char *buf)
{
    int  val;
    int  len;

//    printk("rex5055_WriteWord:\n");

    len = HexToBin(buf, &val);
//    printk("Value: %02X\n", val);

    if (len != -1)
      outw(val, mpv->port+2);       /* Output 16bit data */
    mpv->in_func = NULL;

    return len;
}

/*
 *       "SetBitOn:dd\n"
 */

static int
rex5055_SetBitOn(rexdio_profile *mpv, char *buf)
{
    int  val, set, tmp;
    int  len;

//    printk("rex5055_SetBitOn:\n");

    len = DecToBin(buf, &val);
//    printk("Buf:%s Value: %d\n", buf, val);

    if (len != -1) {
      set = 0x0001 << val;
      tmp = inw(mpv->port+2);
      outw( (tmp | set), mpv->port+2 );
    }
    mpv->in_func = NULL;

//    printk("tmp | set: %04X\n", (tmp | set) );

    return len;
}

/*
 *       "SetBitOff:dd\n"
 */

static int
rex5055_SetBitOff(rexdio_profile *mpv, char *buf)
{
    int  val, set, tmp;
    int  len;

//    printk("rex5055_SetBitOff:\n");
    len = DecToBin(buf, &val);
//    printk("Value: %d\n", val);

    if (len != -1) {
      set = ~( 0x0001 << val );
      tmp = inw(mpv->port+2);
      outw( (tmp & set), mpv->port+2 );
    }
    mpv->in_func = NULL;

//    printk("tmp & set: %04X\n", (tmp & set) );

    return len;

}  /* rex5055_SetBitOff */

/*
 *       "PollIrqStat:\n"
 */

static int
rex5055_PollIrqStat(rexdio_profile *mpv, char *buf)
{

//    printk("rex5055_PollIrqStat:\n");
    mpv->in_func = doIrqStat;
    return 0;
}

/*
 *       "SetIrqResponse:0xDD\n"
 */

static int
rex5055_SetIrqResponse(rexdio_profile *mpv, char *buf)
{
    int  val, set, tmp;
    int  len;

//    printk("rex5055_SetIrqResponse:\n");

    len = HexToBin(buf, &val);
//    printk("Buf:%s Value: %02X\n", buf, val);

    if (len != -1) {
      mpv->res = val;
    }
    mpv->in_func = NULL;

    return len;
}

/*====================================================================*/

static int
doReadMask(rexdio_profile *mpv, char *buf)
{
    int len;

//    printk("doReadMask\n");

    len = sprintf(buf, "M:0x%02X\n", inb(mpv->port+1) );

    mpv->in_func = NULL;
    return len;
}

static int
doReadLowByte(rexdio_profile *mpv, char *buf)
{
    int len;

//    printk("doReadLowByte\n");

    len = sprintf(buf, "L:0x%02X\n", inb(mpv->port+2) );

    mpv->in_func = NULL;
    return len;
}

static int
doReadHighByte(rexdio_profile *mpv, char *buf)
{
    int len;

//    printk("doReadHighByte(0x%X)\n", mpv->port+3);

    len = sprintf(buf, "H:0x%02X\n", inb(mpv->port+3) );

    mpv->in_func = NULL;
    return len;
}

static int
doReadWord(rexdio_profile *mpv, char *buf)
{
    int  len;

//    printk("doReadWord\n");

    len = sprintf(buf, "W:0x%04X\n", inw(mpv->port+2) );

    mpv->in_func = NULL;
    return len;
}

static int
doGetBit(rexdio_profile *mpv, char *buf)
{
    int val, chk, len;
    char *OnOff;

//    printk("doGetBit\n");

    chk = 0x0001 << mpv->bit;
    val = inw(mpv->port+2);

    OnOff = ( val & chk )  ? "On" : "Off";
    len = sprintf(buf, "B:%2d:%s\n", mpv->bit, OnOff);

    mpv->in_func = NULL;
    return len;
}

/*
 * Read routin for "PollIrqStat" command
 */
static int
doIrqStat(rexdio_profile *mpv, char *buf)
{
    int len;

//    printk("doIrqStat\n");

    len = sprintf(buf, "S:0x%02X\n", mpv->stat );

    buf += len;
    switch (mpv->dtype) {
    case 0x03:
        len += sprintf(buf, "W:0x%04X\n", mpv->dat);
        break;
    case 0x01:
        len += sprintf(buf, "L:0x%02X\n", mpv->dat);
        break;
    case 0x02:
        len += sprintf(buf, "H:0x%02X\n", mpv->dat);
        break;
    default:
        len += sprintf(buf, "\n");
        break;
    }
    
    mpv->stat = 0;   /* Clear */
    mpv->in_func = NULL;
    return len;
}

/*====================================================================*/

static ssize_t
rex5055_read(struct file *file, char *buf,
                        size_t count, loff_t * ppos)
{
    rexdio_profile  *mpv;
    int  len;

    mpv = file->private_data;

//    printk("rex5055_read: %d\n", mpv->port);

    if (mpv->in_func == NULL) {
       len = -EAGAIN;
       return len;
    }

    len = mpv->in_func(mpv, buf);

    printk("count:%d: len:%d\n", count, len);

    return len;
}

static ssize_t
rex5055_write(struct file *file, const char *buf,
                                        size_t count, loff_t * ppos)
{
    rexdio_profile  *mpv;
    int  len;

    mpv = file->private_data;
//    printk("rex5055_write:%x count:%d: %p\n", mpv->port, count, buf);

    if ( copy_from_user((char *) mpv->cmdbuf, (char *) buf, count) != 0) {
        printk("copy_from_user error\n");
        return -1;
    }
    *(mpv->cmdbuf + count) = '\0';      /* set end of string */

    len = rex5055_DispatchCmd(mpv, mpv->cmdbuf);
    if ((len != -1) && (*(mpv->cmdbuf+len) == '\n') )
          len++;
//    if (*(mpv->cmdbuf+len+1) == '\n')
//      printk("new line code\n");

//    if (len == -1)
//        printk("Invalid command count=%d / %s\n", count, mpv->cmdbuf);

    return len;
}

static unsigned int rex5055_poll(struct file *file, poll_table * wait)
{
    int mask = 0;

//    printk(" 5055_poll:\n");


    return mask;
}


static int 
rex5055_ioctl(struct inode *inode,
		      struct file *file,
		      unsigned int cmd,
		      unsigned long arg)
{
    int ret;
    rexdio_profile  *mpv;

    mpv = file->private_data;

//    printk(" 5055_ioctl: %04X\n", cmd);

    ret = ioctl5055_DispatchCmd( mpv, cmd, arg);

    return ret;

}  /* rex5055_ioctl */

static int
rex5055_open(struct inode *inode, struct file *file)
{
  int  i;

    printk("rex5055_open: MAJOR:%d PID:%X\n",
            MAJOR(inode->i_rdev), current->pid );

    for (i = 0; i < MPMAX; i++) {
        if (mp[i].major == MAJOR(inode->i_rdev) )
            break;
    }
    if (i == MPMAX)
        return -ENXIO;

    if ((mp[i].pid != 0) && (mp[i].pid != current->pid)) {
        printk("rex5055_open error: MAJOR:%d PID:%X\n",
            MAJOR(inode->i_rdev), current->pid );
        return -EBUSY;    /* already open by other process */
    }

    file->private_data = mp + i;

    mp[i].cmdbuf = kmalloc(256, GFP_KERNEL);
    if ( mp[i].cmdbuf == NULL)
        return -ENOMEM;

    mp[i].stat = 0;
    mp[i].dat = 0;
    mp[i].pid = current->pid;

    switch (MINOR(inode->i_rdev)) {
    case REX5055_MINOR:
        return 0;

    default:
        return -ENXIO;
    }
}

static int rex5055_close(struct inode *inode, struct file *file)
{
    int  i;
    rexdio_profile  *mpv;

    printk("rex5055_close: MAJOR:%d PID:%X\n",
            MAJOR(inode->i_rdev), current->pid );


    mpv = file->private_data;

    if (mpv->pid != current->pid) {
        printk("close error: invalid pid:%X\n", current->pid);
        return -EBUSY;
    }

    kfree(mpv->cmdbuf);

    mpv->pid = 0;

    return 0;
}

/*====================================================================*/

rexdio_ioctl_f   rex_ioctl_tb[] = {
    { REXDIO_SETCTRL, ioctl5055_SetCtrl },        /* 0xD001: Set Control Reg. */
    { REXDIO_SETMASK, ioctl5055_SetMask },        /* 0xD002: Set Mask Reg. */
    { REXDIO_GETMASK, ioctl5055_GetMask },        /* 0xD002: Set Mask Reg. */
    { REXDIO_OUT_HIGHBYTE, ioctl5055_OutHighByte }, /* 0xD010: Out high byte (8-15) */
    { REXDIO_IN_HIGHBYTE, ioctl5055_InHighByte }, /* 0xD011: In high byte (8-15) */
    { REXDIO_OUT_LOWBYTE, ioctl5055_OutLowByte }, /* 0xD012: Out low byte (0-7) */
    { REXDIO_IN_LOWBYTE,  ioctl5055_InLowByte },  /* 0xD013: In low byte (0-7) */
    { REXDIO_OUT_WORD, ioctl5055_OutWord },       /* 0xD014: Out word */
    { REXDIO_IN_WORD, ioctl5055_InWord },         /* 0xD015: In word */
    { REXDIO_IRQSTAT, ioctl5055_IrqStat },        /* 0xD020: Irq status */
    { REXDIO_IRQRESPONSE, ioctl5055_IrqResponse },   /* 0xD021: Irq response */
    { 0, NULL }
} ;

static int
ioctl5055_DispatchCmd(rexdio_profile *mpv,
		      unsigned int cmd,
		      unsigned long arg)
{
    int   i;
    int   ret;

    for (i = 0; rex_ioctl_tb[i].cmd != 0; i++ ) {
        if ( cmd == rex_ioctl_tb[i].cmd ) {
                break;
        }
    }
    if (rex_ioctl_tb[i].cmd == 0) {
        printk("Unknown ioctl command: %04X\n", cmd);
        return -1;
    }


//    printk("Cmd: %s Arg: %X\n", rex_ioctl_tb[i].cmd, arg);

    ret = (rex_ioctl_tb[i].func)(mpv, arg);

//  printk("Cmd: %04X ret: %d\n", rex_ioctl_tb[i].cmd, ret );

    return (ret);

}  /* ioctl5055_DispatchCmd */

/*
 * REXDIO_WRTCTRL 0xD001: Set Control Reg.
 */
static int
ioctl5055_SetCtrl(rexdio_profile *mpv, unsigned long arg)
{
    int val = 0;

//    printk("ioctl5055_SetCtrl(0x%X)\n", mpv->port);

    if ( copy_from_user((char *) &mpv->ctrl, (char *) arg, 1) )
       return -EFAULT;

    outb(mpv->ctrl, mpv->port);

    return 0;
}

/*
 * REXDIO_SETMASK 0xD002: Set Mask Reg.
 */
static int
ioctl5055_SetMask(rexdio_profile *mpv, unsigned long arg)
{
    int val = 0;

//    printk("ioctl5055_SetMask(0x%X)\n", mpv->port+1);

    if ( copy_from_user((char *) &val, (char *) arg, 1) )
       return -EFAULT;

    outb(val, mpv->port+1);

    return 0;
}

/*
 * REXDIO_GETCTRL 0xD003: Set Control Reg.
 */
static int
ioctl5055_GetMask(rexdio_profile *mpv, unsigned long arg)
{
    int val = 0;

    printk("ioctl5055_GetMask(0x%X)\n", mpv->port+1);

    val = inb(mpv->port+1);

    if ( copy_to_user((char *) arg, (char *)&val, 1) )
       return -EFAULT;

    return 0;
}

/*
 * REXDIO_OUT_HIGHBYTE 0xD010: Out high byte (8-15)
 */
static int
ioctl5055_OutHighByte(rexdio_profile *mpv, unsigned long arg)
{
    int val = 0;

    printk("ioctl5055_OutHighByte(0x%X)\n", mpv->port+3);

    if ( copy_from_user((char *) &val, (char *) arg , 1) )
       return -EFAULT;

    outb(val, mpv->port+3);

    return 0;
}

/*
 * REXDIO_IN_HIGHBYTE 0xD011: In high byte (8-15)
 */
static int 
ioctl5055_InHighByte(rexdio_profile *mpv, unsigned long arg)
{
    int val;

    printk("ioctl5055_InHighByte(0x%X)\n", mpv->port+3);

    val = inb(mpv->port+3);

    if ( copy_to_user((char *) arg, (char *)&val, 1) )
       return -EFAULT;

    return 0;
}

/*
 * REXDIO_OUT_LOWBYTE 0xD012: Out low byte (0-7)
 */
static int
ioctl5055_OutLowByte(rexdio_profile *mpv, unsigned long arg)
{
    int val = 0;

    printk("ioctl5055_OutLowByte(0x%X)\n", mpv->port+2);

    if ( copy_from_user((char *) &val, (char *) arg, 1) )
       return -EFAULT;

    outb(val, mpv->port+2);

    return 0;
}
/*
 * REXDIO_IN_LOWBYTE 0xD013: In low byte (0-7)
 */
static int
ioctl5055_InLowByte(rexdio_profile *mpv, unsigned long arg)
{
    int val;

    printk("ioctl5055_InLowByte(0x%X)\n", mpv->port+2);

    val = inb(mpv->port+2);

    if ( copy_to_user((char *) arg, (char *) &val, 1) )
       return -EFAULT;

    return 0;
}

/*
 * REXDIO_OUT_WORD 0xD014: Out word
 */
static int
ioctl5055_OutWord(rexdio_profile *mpv, unsigned long arg)
{
    int val;

    printk("ioctl5055_InOutWord(0x%X)\n", mpv->port+2);

    if ( copy_from_user((char *) &val, (char *) arg, 2) )
       return -EFAULT;

    outw(val, mpv->port+2);

    return 0;
}

/*
 * REXDIO_IN_WORD 0xD015: In word
 */
static int
ioctl5055_InWord(rexdio_profile *mpv, unsigned long arg)
{
    int val;

    printk("ioctl5055_InWord(0x%X)\n", mpv->port+2);

    val = inw(mpv->port+2);

    if ( copy_to_user((char *) arg, (char *) &val, 2) )
       return -EFAULT;

    return 0;

}  /* ioctl5055_InWord */

/*
 * REXDIO_IRQSTAT 0xD020: Poll Irq status
 */
static int
ioctl5055_IrqStat(rexdio_profile *mpv, unsigned long arg)
{
    unsigned char  *parg;

//    printk("ioctl5055_IrqStat(0x%X, 0x%X)\n",
//            mpv->stat, mpv->dat);

    parg = (char *)arg;

    if ( copy_to_user(parg, (char *) &mpv->stat, 2) )
       return -EFAULT;
    if ( copy_to_user(parg + 2, (char *) &mpv->dat, 2) )
       return -EFAULT;

    mpv->stat = 0;   /* Clear */

    return 0;
}  /* ioctl5055_IrqStat */

/*
 * REXDIO_IRQRESPONSE 0xD021: Set Irq response
 */
static int 
ioctl5055_IrqResponse(rexdio_profile *mpv, unsigned long arg)
{
    printk("ioctl5055_IrqResponse(0x%X)\n", mpv->stat);

    if ( copy_from_user((char *) &mpv->res, (char *) arg, 2) )
       return -EFAULT;

    return 0;

}  /* ioctl5055_IrqResponse */

/*====================================================================*/

static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}
/*======================================================================

    serial_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/

static dev_link_t *rex5055_attach(void)
{
    rex5055_info_t *info;
    client_reg_t client_reg;
    dev_link_t *link;
    int i, ret;

//    printk("rex5055_attach:\n");
    DEBUG(0, "rex5055_attach()\n");

    /* Create new serial device */
    info = kmalloc(sizeof(*info), GFP_KERNEL);
    if (!info) return NULL;
    memset(info, 0, sizeof(*info));
    info->Socket = -1;	/* mp is not assigned */
    link = &info->link;
    link->priv = info;

    link->release.function = &rex5055_release;
    link->release.data = (u_long)link;
    /* REX-5055 is 16bit access mode */
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
    link->io.NumPorts1 = 4;
//    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE | IRQ_HANDLE_PRESENT;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    if (irq_list[0] == -1)
        link->irq.IRQInfo2 = irq_mask;
    else
        for (i = 0; i < 4; i++)
            link->irq.IRQInfo2 |= 1 << irq_list[i];

    link->irq.Handler = &rex5055_interrupt;

    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;

    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
        CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
        CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
        CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &rex5055_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != CS_SUCCESS) {
        cs_error(link->handle, RegisterClient, ret);
        rex5055_detach(link);
        return NULL;
    }

    return link;
} /* rex5055_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void rex5055_detach(dev_link_t *link)
{
    rex5055_info_t *info = link->priv;
    dev_link_t **linkp;
    int ret;

//    printk("rex5055_dettach:\n");
    DEBUG(0, "rex5055_detach(0x%p)\n", link);

    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
        if (*linkp == link) break;

    if (*linkp == NULL)
        return;

    del_timer(&link->release);
//    if (link->state & DEV_CONFIG)
//        rex5055_release((u_long)link);

    if (link->handle) {
	ret = CardServices(DeregisterClient, link->handle);
	if (ret != CS_SUCCESS)
	    cs_error(link->handle, DeregisterClient, ret);
    }

    /* Unlink device structure, free bits */
    *linkp = link->next;
    kfree(info);

} /* rex5055_detach */

/*====================================================================*/

static int get_tuple(int fn, client_handle_t handle, tuple_t *tuple,
		     cisparse_t *parse)
{
    int i;
    i = CardServices(fn, handle, tuple);
    if (i != CS_SUCCESS) return CS_NO_MORE_ITEMS;
    i = CardServices(GetTupleData, handle, tuple);
    if (i != CS_SUCCESS) return i;
    return CardServices(ParseTuple, handle, tuple, parse);
}

/*======================================================================

    serial_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    serial device available to the system.

======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn), args))!=0) goto cs_failed

#define CFG_CHECK(fn, args...) \
if (CardServices(fn, args) != 0) goto next_entry

static void
rex5055_config(dev_link_t *link)
{
    client_handle_t handle = link->handle;
    rex5055_info_t *info = link->priv;
    tuple_t tuple;
    u_short buf[128];
    cisparse_t parse;
    cistpl_cftable_entry_t *cfg = &parse.cftable_entry;
    cistpl_cftable_entry_t dflt = { 0 };
    int last_ret, last_fn;
    char  dev_name[9];

//    printk("rex5055_config(0x%p)\n", link);
    DEBUG(0, "rex5055_config(0x%p)\n", link);

    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleOffset = 0;
    tuple.TupleDataMax = 255;
    tuple.Attributes = 0;
    /* Get configuration register information */
    tuple.DesiredTuple = CISTPL_CONFIG;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* Configure card */
//    link->state |= DEV_CONFIG;

    /* Is this a compliant multifunction card? */
//    tuple.DesiredTuple = CISTPL_LONGLINK_MFC;
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
    tuple.Attributes = 0;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    while (1) {
        CFG_CHECK(GetTupleData, handle, &tuple);
        CFG_CHECK(ParseTuple, handle, &tuple, &parse);
        if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0)) {
            cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
            link->conf.ConfigIndex = cfg->index;
            link->io.BasePort1 = io->win[0].base;
            link->io.NumPorts1 = io->win[0].len;
            link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
            CFG_CHECK(RequestIO, link->handle, &link->io);
            /* If we've got this far, we're done */
            break;
        }

    next_entry:
        if (cfg->flags & CISTPL_CFTABLE_DEFAULT) dflt = *cfg;
        CS_CHECK(GetNextTuple, handle, &tuple);
    }

    CS_CHECK(RequestIRQ, handle, &link->irq);
    CS_CHECK(RequestConfiguration, handle, &link->conf);

//    if (info->ndev == 0)
//        goto failed;

    mp_socket = handle->Socket;
    mp[mp_socket].socket = mp_socket;
    mp[mp_socket].port = link->io.BasePort1;
    mp[mp_socket].irq = link->irq.AssignedIRQ;
    info->ndev = 1;
    sprintf(dev_name, "REX5055");
    printk("rex5055d: Socket:%d IO=0x%X IRQ=%d ",
           mp_socket, mp[mp_socket].port, mp[mp_socket].irq);
    rex5055_major = devfs_register_chrdev(mp[mp_socket].major,
                                          dev_name, &rex5055_fops);
    if (mp[mp_socket].major == 0) {
        if (rex5055_major == 0) {
            printk(KERN_ERR "\nREX5055 DIO - cannot register device %d\n",
                        mp[mp_socket].major);
            goto failed;
        }
        mp[mp_socket].major = rex5055_major;
    }

    mp[mp_socket].pid = 0;   /* not used */

    if (rex5055_dev_probe() == 0)
        printk(", MAJOR %d Name:%s\n", mp[mp_socket].major, dev_name);
    devfs_handle = devfs_register (NULL, dev_name, DEVFS_FL_DEFAULT,
                                            mp[mp_socket].major, REX5055_MINOR,
                                            S_IFCHR | S_IRUSR | S_IWUSR,
                                            &rex5055_fops, NULL);

    mp[mp_socket].handle = devfs_handle;
    info->Socket = mp_socket;

    link->dev = &info->node[0];

    return;

cs_failed:
    cs_error(link->handle, last_fn, last_ret);
failed:
    rex5055_release((u_long)link);

} /* rex5055_config */

/*======================================================================

    After a card is removed, serial_release() will unregister the net
    device, and release the PCMCIA configuration.

======================================================================*/

void rex5055_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    rex5055_info_t *info = link->priv;
    rexdio_profile *mpv;

//    printk("rex5055_release(0x%p)\n");
    DEBUG(0, "rex5055_release(0x%p)\n", link);

/*
    if (mpv->cmdbuf != NULL) {
        kfree(mpv->cmdbuf);
        mpv->cmdbuf = NULL;
    }
*/

    CardServices(ReleaseConfiguration, link->handle, &link->conf);
    CardServices(ReleaseIRQ, link->handle, &link->irq);
    CardServices(ReleaseIO, link->handle, &link->io);

    if (info->Socket != -1) {
      mpv = mp + info->Socket;
      devfs_unregister(mpv->handle);
      devfs_unregister_chrdev(mpv->major, "REX5055");
      mpv->handle = NULL;   /* clear mp */
// printk("release: Socket %d MAJOR %d\n," , info->Socket, mpv->major);
    }
//    mp_count--;

    link->dev = NULL;
//    link->state &= ~DEV_CONFIG;

} /* rex5055_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the serial drivers from
    talking to the ports.

======================================================================*/

static int rex5055_event(event_t event, int priority,
			event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;
    rex5055_info_t *info = link->priv;

//    printk("rex5055_event(0x%06X)\n", event);
    DEBUG(1, "rex5055_event(0x%06x)\n", event);

    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
//        link->state &= ~DEV_PRESENT;
//        if (link->state & DEV_CONFIG)
            mod_timer(&link->release, jiffies + HZ/20);
        break;
    case CS_EVENT_CARD_INSERTION:
//        link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
        rex5055_config(link);
        break;
    case CS_EVENT_PM_SUSPEND:
//        link->state |= DEV_SUSPEND;
        /* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
//        if (link->state & DEV_CONFIG)
            CardServices(ReleaseConfiguration, link->handle);
        break;
    case CS_EVENT_PM_RESUME:
//        link->state &= ~DEV_SUSPEND;
        /* Fall through... */
    case CS_EVENT_CARD_RESET:
        if (DEV_OK(link) )
            CardServices(RequestConfiguration, link->handle, &link->conf);
        break;
    }
    return 0;
} /* rex5055_event */

/* ------------------------------------------------------------------------ */

static void
rex5055_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  rexdio_profile *mpv;
  int  i;
  int  st;

//    printk("rex5055_interrupt: irq:%X %p\n", irq, dev_id);

    for (i = 0; i < MPMAX; i++) {
        if (mp[i].irq == irq)
            break;
    }
    if (i == MPMAX)
        return;    /* not found */

    mpv = mp+i;
    mpv->stat = inb(mpv->port+1);	/* Get mask data */

    if (mpv->stat == 0xff) {
        printk("Card removed: stat:%X\n", mpv->stat);
        return;
    }

    if ((mpv->stat & 0x80) == 0) {
        printk("No interrupt: stat:%X\n", mpv->stat);
        return;    /* no interrupt */
    }
    mpv->dtype = mpv->stat & 0x03;

    switch (mpv->dtype) {
    case 0x03:
        mpv->dat = inw(mpv->port+2);
        break;
    case 0x01:
        mpv->dat = inb(mpv->port+2);
        break;
    case 0x02:
        mpv->dat = inb(mpv->port+3);
        break;
    }

    st = mpv->stat & 0xC0;
    switch (st) {
    case 0x0C:
        outw(mpv->res, mpv->port+2);
        break;
    case 0x01:
        outb(mpv->res, mpv->port+2);
        break;
    case 0x02:
        outb(mpv->res, mpv->port+3);
        break;
    }

//    printk("%04X: %02X %04X\n", mpv->port, mpv->stat, mpv->dat );

    return;
}


/* ------------------------------------------------------------------------ */

static int __init rex5055_init(void)
{
    int  ret;
    int  i;
    servinfo_t serv;

//    printk("rex5055_init: ");
    printk("%s\n", version);
    DEBUG(0, "%s\n", version);
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) {
	printk(KERN_NOTICE "rex5055d: Card Services release "
	       "does not match!\n");
	return -1;
    }
    for (i = 0; i < MPMAX; i++) {
        mp[i].irq = -1;
        mp[i].major = 0;
        mp[i].handle = NULL;    /* clear mp */
    }
    ret = register_pccard_driver(&dev_info, &rex5055_attach, &rex5055_detach);

    MOD_INC_USE_COUNT;

    return 0;

}

static void __exit rex5055_cleanup (void)
{
//    int  i;
//    char  dev_name[9];

//    printk("rex5055_cleanup:\n");

//    current->state = TASK_INTERRUPTIBLE;

    DEBUG(0, "rex5055d: unloading\n");
    unregister_pccard_driver(&dev_info);

/*
    for (i = mp_count-1; i >= 0; i--) {
//      sprintf(dev_name, "REX5055%d", i);
      sprintf(dev_name, "REX5055");
      devfs_unregister_chrdev(mp[i].major, dev_name);
      devfs_unregister(mp[i].handle);
    }
*/

    MOD_DEC_USE_COUNT;

}

module_init(rex5055_init);
module_exit(rex5055_cleanup);

/* ------------------------------------------------------------------------ */


static int
__init rex5055_dev_probe(void)
{
    unsigned char val0, val1, val2, val3;

//    printk("rex5055_dev_probe:\n");
/*
    val0 = inb(mp.port);
    val1 = inb(mp.port+1);
    val2 = inb(mp.port+2);
    val3 = inb(mp.port+3);
    printk("rex5055_dev_probe: %02X %02X %02X %02X\n",
                val0, val1, val2, val3);
*/
    return 0;
}

